from Acquisition import ImplicitAcquisitionWrapper, aq_parent
from Products.CMFCore.utils import getToolByName
from Products.Archetypes.public import *
from Products.Archetypes.ReferenceEngine import Reference
from Products.Archetypes.utils import make_uuid

from Products.Relations.config import *
from Products.Relations import interfaces, ruleset, utils
from Products.Relations.schema import BaseSchemaWithInvisibleId

class InverseImplicator(BaseContent, ruleset.RuleBase):
    """Implicator that creates a reference from target to source."""
    __implements__ = (interfaces.IImplicator,) + BaseContent.__implements__

    content_icon = 'inverseimplicator_icon.gif'

    def implyOnConnect(self, reference, chain):
        iRuleset = self.getInverseRuleset()
        if iRuleset is not None:
            iRuleset.implyOnConnect(reference.getTargetBrain(),
                                    reference.getSourceBrain(),
                                    chain)

    def implyOnDisconnect(self, reference, chain):
        iRuleset = self.getInverseRuleset()
        if iRuleset is None:
            return

        rc = getToolByName(self.getRuleset(), REFERENCE_CATALOG)
        result = rc(sourceUID=reference.targetUID,
                    targetUID=reference.sourceUID,
                    relationship=iRuleset.getId())
        if result:
            iRuleset.implyOnDisconnect(result[0].getObject(), chain)

    # Archetypes implementation
    schema = BaseSchemaWithInvisibleId + Schema((
        ReferenceField('inverseRuleset',
                       relationship='inverseRuleset',
                       allowed_types_method='_allowedTypesRuleset'),
        ))
    portal_type = 'Inverse Implicator'

    def _allowedTypesRuleset(self, instance):
        return utils.getPortalTypesByInterfaces(self, ('IRuleset',))

    def getInverseRuleset(self):
        inv = self.getRefs('inverseRuleset')
        if inv:
            return inv[0]
        else:
            return None

registerType(InverseImplicator)
                       
        

    
