## Script (Python) "relations_adddelete_vocab.py"
##parameters=test_only=0, ruleset_ids=None
##title=Make Relations vocabularies by querying all rulesets.
##bind context=context
##bind container=container
##
from Products.Relations.brain import makeBrainAggregate
from Products.Relations.utils import isReferenceable

# What we return here is a list of dictionaries with the following keys:
#     - id      the ruleset's id, which is the relationship
#     - title   the ruleset's title
#     - tuples  a list of tuples, see below.
#
# Each 'tuples' item is a 2-tuple in the form (aggregated brain, selected),
# where selected is true if there's a reference
# (source=context, target=vocab item, relationship=ruleset.id).
#
# If test_only=1, I will return true if *any* vocabulary item exists. Otherwise
# I will return a false value.

value = []

if not isReferenceable(context):
    return []

library = context.relations_library
refctl = context.reference_catalog
mtool = context.portal_membership

if not ruleset_ids:
    rulesets = library.getRulesets()
else:
    rulesets = [library.getRuleset(rid) for rid in ruleset_ids]

for ruleset in rulesets:
    if not mtool.checkPermission('View', ruleset): continue
    relationship = ruleset.getId()
    vocabulary = ruleset.makeVocabulary(context)
    if vocabulary is None: vocabulary = []
    refs = refctl(sourceUID=context.UID(), relationship=relationship)
    existing_targetUIDs = [r.targetUID for r in refs]
    if not vocabulary and not existing_targetUIDs:
        continue # we don't add an entry
    elif test_only:
        return True

    entry = {'id': relationship,
             'title': ruleset.Title(),
             'tuples': []}
    value.append(entry)

    tuples = entry['tuples']

    # first add all vocabulary items to 'tuples' in entry dict
    for item in vocabulary:
        selected = False
        if item.UID in existing_targetUIDs:
            selected = True
        tuples.append((item, selected))

    # now find existing refs that are not part of the vocab, which we add
    vocab_uids = [item.UID for (item, selected) in tuples]
    for uid in existing_targetUIDs:
        if uid not in vocab_uids:
            tuples.append((makeBrainAggregate(context, uid), True))

return value
        
    
